package edu.ntnu.stud.chaosgame.controller.game;

import edu.ntnu.stud.chaosgame.controller.utility.ChaosGameFileHandler;
import edu.ntnu.stud.chaosgame.controller.utility.PopupManager;
import edu.ntnu.stud.chaosgame.model.game.ChaosCanvas;
import edu.ntnu.stud.chaosgame.model.game.ChaosGameDescription;
import edu.ntnu.stud.chaosgame.model.generators.ChaosGameDescriptionFactory;
import edu.ntnu.stud.chaosgame.model.transformations.JuliaTransform;
import edu.ntnu.stud.chaosgame.view.ChaosCanvasToImageConverter;
import edu.ntnu.stud.chaosgame.view.ChaosGameGui;
import edu.ntnu.stud.chaosgame.view.modificationpopups.Affine2DPopup;
import edu.ntnu.stud.chaosgame.view.modificationpopups.JuliaPopup;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.ComboBox;
import javafx.scene.image.PixelReader;
import javafx.scene.image.WritableImage;
import javafx.scene.paint.Color;
import javafx.stage.Popup;
import javafx.util.Duration;
import javafx.stage.FileChooser;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;

/**
 * Controller class for the GUI buttons, responsible for handling their functionality.
 */
public class GuiButtonController {

  /**
   * The file handler for the controller.
   */
  private ChaosGameFileHandler fileHandler;

  /**
   * The chaos game for the controller.
   */
  public ChaosGame game;

  /**
   * The GUI for the controller.
   */
  private ChaosGameGui gui;

  /**
   * The timeline for the controller.
   */
  private Timeline timeline;

  /**
   * The ChaosCanvas for the controller.
   */
  private ChaosCanvas chaosCanvas;

  /**
   * The factory for the ChaosGameDescriptions.
   */
  private ChaosGameDescriptionFactory factory;

  /**
   * The controller which assigns functionality to the popup-window
   */
  private PopupButtonController popupButtonController;

  /**
   * Default image save path
   */
  private String currentImagePath = System.getProperty("user.home");

  /**
   * The step counter for the controller.
   */
  private int stepCounter = 0;


    /**
   * Basic constructor for the GuiButtonController.
   *
   * @param game the chaos game.
   * @param gui the GUI.
   */
    public GuiButtonController(ChaosGame game, ChaosGameGui gui) {

      this.game = game;
      this.gui = gui;
      this.timeline = new Timeline(new KeyFrame(Duration.seconds(0.05), event -> this.drawChaosGame()));
      this.timeline.setCycleCount(Timeline.INDEFINITE);
      this.factory = new ChaosGameDescriptionFactory();
      this.chaosCanvas = new ChaosCanvas(1000, 1000, this.factory.getDescriptions().get(0).getMinCoords(),
              this.factory.getDescriptions().get(0).getMaxCoords());
      initializeMainButtons();
      this.fileHandler = new ChaosGameFileHandler();
      this.popupButtonController = new PopupButtonController();
      initializeLoadFractalFromFileButtonHandler();
      initializeWriteToFileButtonHandler();
      initializeModifyGameButtonHandler();
      initializeDescriptionComboBox();
      initializeSaveImageButtons();
    }

  /**
   * Initialize the modify game button handler.
   */
  private void initializeModifyGameButtonHandler() {
    gui.getModifyGameButton().setOnAction(e -> {
      ChaosGameDescription description = game.getDescription();
      if (description.getTransforms().get(0) instanceof JuliaTransform) {
        JuliaPopup juliaPopup = new JuliaPopup(description);
        juliaPopup.getPopupModifyStage().setUserData(this);
        popupButtonController.handleUpdateButton(juliaPopup);
        juliaPopup.display();
      } else {
        Affine2DPopup affine2DPopup = new Affine2DPopup(description);
        affine2DPopup.getPopupModifyStage().setUserData(this);
        popupButtonController.handleUpdateButton(affine2DPopup);
        affine2DPopup.display();
      }
    });
  }

  /**
   * Initialize the description combo box.
   */
  private void initializeDescriptionComboBox() {
    ComboBox<String> descriptionComboBox = gui.getDescriptionComboBox();

    descriptionComboBox.setOnShowing(event -> updateDescriptionComboBox());

    descriptionComboBox.getSelectionModel().selectedItemProperty()
      .addListener((observable, oldValue, newValue) -> {
        for (int i = 0; i < factory.getDescriptions().size(); i++) {
          ChaosGameDescription description = factory.getDescriptions().get(i);
          if (description.getName().equals(newValue)) {
            updateDescription(i);
            break;
          }
        }

      });
    }

    /*
      * Update the description combo box.
     */
    public void updateDescriptionComboBox() {
      factory = new ChaosGameDescriptionFactory();
      ComboBox<String> descriptionComboBox = gui.getDescriptionComboBox();
      descriptionComboBox.getItems().clear();
      for (ChaosGameDescription description : factory.getDescriptions()) {
          descriptionComboBox.getItems().add(description.getName());
      }
    }

  /**
   * Initialize the save image buttons.
   */
  private void initializeSaveImageButtons(){
      gui.getSaveImageButton().setOnAction(event -> saveImage());
    }

  /**
   * Initialize the load fractal from file button handler.
   */
  private void initializeLoadFractalFromFileButtonHandler() {
    gui.getLoadFractalFromFileButton().setOnAction(event -> {
      FileChooser fileChooser = new FileChooser();
      fileChooser.setTitle("Open Fractal File");
      fileChooser.getExtensionFilters().add(
              new FileChooser.ExtensionFilter("Text Files", "*.txt"));
      File selectedFile = fileChooser.showOpenDialog(gui.getStage());
      if (selectedFile != null) {
        Path dest = Paths.get("src/main/resources/descriptions/saved_descriptions/"
            + selectedFile.getName());
        try {
          Files.copy(selectedFile.toPath(), dest);
          // Refresh the list of descriptions
          factory = new ChaosGameDescriptionFactory();
        } catch (IOException e) {
          PopupManager.displayError("Error", "Could not load the fractal from file.");
        }
      }
    });
  }

  /**
   * Initialize the write to file button handler.
   */
  private void initializeWriteToFileButtonHandler() {
    gui.getWriteToFileButton().setOnAction(new EventHandler<ActionEvent>() {
      @Override
      public void handle(ActionEvent event) {
        // Create a FileChooser object
        FileChooser fileChooser = new FileChooser();

        // Set the title of the FileChooser dialog
        fileChooser.setTitle("Save File");

        // Set the initial directory of the FileChooser
        fileChooser.setInitialDirectory(new File(System.getProperty("user.home")));

        // Set the extension filter
        fileChooser.getExtensionFilters().addAll(
          new FileChooser.ExtensionFilter("Text Files", "*.txt"),
          new FileChooser.ExtensionFilter("All Files", "*.*")
        );

        // Show the save file dialog and get the selected file
        File file = fileChooser.showSaveDialog(gui.getStage());

        if (file != null) {
          try {
            // Create a ChaosGameDescription object
            ChaosGameDescription description = game.getDescription();

            // Call the writeToFile method
            fileHandler.writeToFile(description, file.getPath());
          } catch (IOException e) {
            // Handle the exception
            PopupManager.displayError("Error", "Could not write the ChaosGameDescription to file.");
          }
        }
        }
      });
    this.timeline = new Timeline(new KeyFrame(Duration.seconds(0.05), event -> this.drawChaosGame()));
    this.timeline.setCycleCount(Timeline.INDEFINITE);
    this.factory = new ChaosGameDescriptionFactory();
    this.chaosCanvas = new ChaosCanvas(1000, 1000, this.factory.getDescriptions().get(0).getMinCoords(),
            this.factory.getDescriptions().get(0).getMaxCoords());
    initializeMainButtons();
  }

  /**
   * Draw the chaos game on the canvas.
   */
  public void drawChaosGame(){
    int iterationLimit = Integer.parseInt(gui.getIterationLimitTextField().getText());

    if (stepCounter >= iterationLimit) {
      timeline.stop();
      return;
    }
    // Run the number of steps specified in text field, else 1000.
    game.runSteps(!Objects.equals(gui.getStepCountTextField().getText(), "") ?
        Integer.parseInt(gui.getStepCountTextField().getText()) : 1000);

    // Convert the canvas to either an image with coloured pixels based on intensity, or just black and white.
    ChaosCanvasToImageConverter converter = new ChaosCanvasToImageConverter(chaosCanvas,
        gui.getColorCheckBox().isSelected());

    WritableImage image = converter.getImage();
    gui.getCanvas().getGraphicsContext2D().drawImage(image, 0, 0);
    gui.getImageView().setImage(image);
      stepCounter++;
  }

    /**
     * Saves a WritableImage to an image file.
     */
    public void saveImage() {
      // Retrieve the WritableImage from the ImageView
      WritableImage image = gui.getImageView().snapshot(null, null);

      // Use a FileChooser to let the user specify the filename
      FileChooser fileChooser = new FileChooser();
      fileChooser.setTitle("Save Image");
      fileChooser.setInitialDirectory(new File(currentImagePath));
      fileChooser.getExtensionFilters().add(new FileChooser.ExtensionFilter("PNG Files", "*.png"));
      File file = fileChooser.showSaveDialog(gui.getStage());

      if (file != null) {
        try {
          BufferedImage bufferedImage = toBufferedImage(image);
          ImageIO.write(bufferedImage, "png", file);
        } catch (IOException e) {
          PopupManager.displayError("Save error", "There was an error whilst saving the image.");
        }
      }
    }

  /**
   * Converts a WritableImage to a BufferedImage.
   *
   * @param image the WritableImage to convert.
   * @return the converted BufferedImage.
   */
  private BufferedImage toBufferedImage(WritableImage image) {
    int width = (int) image.getWidth();
    int height = (int) image.getHeight();
    BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
    PixelReader reader = image.getPixelReader();
    for (int y = 0; y < height; y++) {
      for (int x = 0; x < width; x++) {
        Color color = reader.getColor(x, y);
        int r = (int) (color.getRed() * 255);
        int g = (int) (color.getGreen() * 255);
        int b = (int) (color.getBlue() * 255);
        int a = (int) (color.getOpacity() * 255);
        int argb = (a << 24) | (r << 16) | (g << 8) | b;
        bufferedImage.setRGB(x, y, argb);
      }
    }
    return bufferedImage;
  }

  /**
   * Start the game.
   */
  public void startGame() {
    timeline.play();
    gui.getStartButton().setText("Start");
  }

  /**
   * Stop the game.
   */
  public void stopGame() {
    timeline.stop();
    gui.getStartButton().setText("Resume");
  }

  /**
   * Clear the image view.
   */
  public void clearCanvas() {
    chaosCanvas.clearCanvas();
    GraphicsContext gc = gui.getCanvas().getGraphicsContext2D();
    gc.clearRect(0, 0, gui.getCanvas().getWidth(), gui.getCanvas().getHeight());
    gui.getImageView().setImage(null);
  }


  /**
   * Quit the game.
   */
  public void quitGame() {
    Platform.exit();
  }

  /**
   * Initialize the main buttons.
   */
  private void initializeMainButtons() {
    gui.getStartButton().setOnAction(event -> startGame());
    gui.getStopButton().setOnAction(event -> stopGame());

    gui.getClearButton().setOnAction(event -> clearCanvas());

    gui.getQuitButton().setOnAction(event -> quitGame());
    gui.getColorCheckBox().setOnAction(event -> game.setUseColor(gui.getColorCheckBox().isSelected()));

  }

  /**
   * Update the description of the chaos game.
   * TODO: this method may need to be changed depending on how we implement the UI. Rename to update?
   *
   * @param index the index of the new description in the list of factory descriptions.
   *
   */
  public void updateDescription(int index) {
    // Stop the timeline
    timeline.stop();

    // Clear the ChaosCanvas
    chaosCanvas.clearCanvas();
    gui.getCanvas().getGraphicsContext2D().clearRect(0, 0, gui.getCanvas().getGraphicsContext2D().
            getCanvas().getWidth(), gui.getCanvas().getGraphicsContext2D().getCanvas().getHeight());
    clearCanvas();
    chaosCanvas.clearCanvas();

    // Get the new description and create a new ChaosCanvas
    ChaosGameDescription description = factory.getDescriptions().get(index);
    chaosCanvas = new ChaosCanvas(1000, 1000, description.getMinCoords(),
            description.getMaxCoords());

    // Update the ChaosCanvas in the GUI
    gui.updateCanvas(chaosCanvas);

    // Create a new ChaosGame with the updated description and ChaosCanvas
    game = new ChaosGame(description, chaosCanvas);
    game.setUseColor(gui.getColorCheckBox().isSelected());
  }
}